'''
Module: pmodels.lin
-------------------

* Model: LIN -> system property = linear combination of component properties.
* Systems: an arbitrary binary polymer systems (or any binary system).
* Predictions: an arbitrary property, but just the first approximation.

Simple usage

>>> # Standard import of pmodels package
>>> import pmodels as pm
>>>
>>> # (1) Estimate elastic modulus (E) of an isotropic binary polymer blend,
>>> # where moduli of the components are E1 = 1 GPa and E2 = 3 GPa,
>>> # and volume fraction of the second component is 0.5.
>>> E_lin = pm.lin.P(P1=1, P2=3, v2=0.5)
>>>
>>> # (2) Print the result
>>> print(f'Modulus of the blend: {E_lin:.2f}')
Modulus of the blend: 2.00

Notes

* LIN is a very simple model ~ rule of mixtures.
* It is equivalent to Voigt rule (pmodels.comp.VoigtRule.P).
* Both LIN model and Voigt rule are based on iso-strain assumption.
    - they are used mostly for predicting elastic modulus
    - but other properties (mech.props, conductivity) can be estimated as well
* In the field of polymers, LIN model works well for:
    - E of polymer composites with (infinitely) long oriented fibers
    - H of semicryscalline polymers (amorph. and cryst. phases interconnected)
    - and analogous systems, where iso-strain assumption holds
'''

def P(P1,P2,v2):
    '''
    Calculate (arbitrary) property P using linear model.
    
    Parameters
    ----------
    P1 : float
        Property of the 1st component.
    P2 : float
        Property of the 2nd component.
    v2 : float
        Volume fraction of the 2nd component.
            
    Returns
    -------
    P : float
        P = final property of the system according to linear model.
        
    Test
    ----
    >>> P(100,200,v2=0.5)
    150.0
    '''
    P = P1*(1-v2) + P2*v2
    return(P)



'''
Module: pmodels.comp
--------------------

* Collection of various predictive models for polymer composites.
* Mostly for two-component composites, isotropic, with isometric filler.
* The models can predict elastic modulus (E) or an arbitrary property (P).

Simple usage

>>> # Standard import of pmodels package
>>> import pmodels as pm
>>>
>>> # (1) Estimate elastic modulus (E) of a polymer composite,
>>> # with an epoxy matrix (E1 ~ 2.5 GPa),
>>> # and 5% of silica particles (E2 ~ 70 GPa).
>>> E_voigt = pm.comp.VoigtRule.P(P1=2.5, P2=70, v2=0.05)
>>> E_reuss = pm.comp.ReussRule.P(P1=2.5, P2=70, v2=0.05)
>>> E_guth  = pm.comp.Guth.E(Em=2.5, vf=0.05)
>>>
>>> # (2) Print the results
>>> # (Note: experimental value ~ 3.0 GPa
>>> print(f'Voigt: {E_voigt:.2f}, Reuss: {E_reuss:.2f}, Guth: {E_guth:.2f}')
Voigt: 5.88, Reuss: 2.63, Guth: 2.90
'''

import sys


class VoigtRule:
    '''
    Voigt rule for binary polymer systems.
    
    * Voigt rule = iso-strain assumption
      = the strain is uniform accross all components.
    * The property is a *weighted arithmetic mean* of the constituent props.
    * Voigt and Reuss rule are bounds,
      real composite properties usually fall between them.
    * Voigt rule is the upper bound, Reuss rule is the lower bound.
    * Voigt rule ~ rule of mixtures, Reuss rule ~ inverse rule of mixtures.
    
    Note
    
    * Voigt rule is equivalent to LIN model.
    * See more details in notes to pmodels.lin module.
    '''
    
    def P(P1,P2,v2):
        '''
        Estimate composite property by means of Voigt rule = upper bound.
        
        * In typical case, the Voigt rule
          predicts upper bound of elastic modulus, E.

        Parameters
        ----------
        P1 : float
            The property of the 1st component (usually the matrix).
        P2 : float
            The property of the 2nd component (usually the filler).
        v2 : float
            The volume fraction of the 2nd component.
            The volume fraction of the 1st component
            can be calculated as v1 = (1 - v2).

        Returns
        -------
        P : float
            Property of the composite acc.to Voight Rule.
        
        Test
        ----
        >>> print(f'{VoigtRule.P(P1=1, P2=100, v2=0.05):.2f}')
        5.95
        '''
        v1 = 1 - v2
        P = P1*v1 + P2*v2
        return(P)


class ReussRule:
    '''
    Reuss rule for binary polymer systems.
    
    * Reuss rule = iso-stress assumption
      = the stress is uniform accross all components.
    * The property is a *weighted harmonic mean* of the constituent props.
    * Voigt and Reuss rule are bounds,
      real composite properties usually fall between them.
    * Reuss rule is the upper bound, Reuss rule is the lower bound.
    * Voigt rule ~ rule of mixtures, Reuss rule ~ inverse rule of mixtures.
    '''
    
    def P(P1,P2,v2):
        '''
        Estimate composite property by means of Reuss rule = lower bound.
        
        * In typical case, the Reuss rule
          predicts lower bound of elastic modulus, E.

        Parameters
        ----------
        P1 : float
            The property of the 1st component (usually the matrix).
        P2 : float
            The property of the 2nd component (usually the filler).
        v2 : float
            The volume fraction of the 2nd component.
            The volume fraction of the 1st component
            can be calculated as v1 = 1 = v2.
            
        Returns
        -------
        P : float
            Property of the composite acc.to Voight Rule.
        
        Test
        ----
        >>> print(f'{ReussRule.P(P1=1, P2=100, v2=0.05):.2f}')
        1.05
        '''

        v1 = 1 - v2
        P = 1 / (v1/P1 + v2/P2)
        return(P)
    

class Einstein:
    '''
    Einstein equation for polymer composites = matrix with filler.
    
    * Einstein equation takes two forms:
        - for bad interfacial adhesion:  Ec = Em(1 + vf)
        - for good interfacial adhesion: Ec = Em(1 + 2.5*vf)
    * Assumption #1 (for both forms of Einstein eq.): Ef >> Em,
      i.e. the filler is much stiffer than the matrix.
    * Assumption #2 (for both forms of Einstein eq.): vf <= 0.05,
      i.e. the filler content is low (below 5 vol.%).
    '''

    def E(Em,vf, adhesion = 0):
        '''
        Predict elastic modulus of a composite by means of Einstein equation.

        Parameters
        ----------
        Em : float
            Elastic modulus of the polymer matrix.
        vf : float
            Volume fraction of the filler.
        adhesion : 0 or 1, optional, default is 1
            If adhesion = 0
            => negligible interfacial adhesion => Ec = Em * (1 + vf).
            If adhesion = 1
            => perfect interfacial adhesion => Ec = Em * (1 + 2.5*vf).

        Returns
        -------
        E : float
            Elastic modulus of the composite.
            
        Test
        ----
        >>> print(f'{Einstein.E(1, 0.05, adhesion=0):.3f}')
        1.050
        >>> print(f'{Einstein.E(1, 0.05, adhesion=1):.3f}')
        1.125
        '''
        if adhesion == 0:
            return( Em * (1 + vf)     )
        elif adhesion == 1:
            return( Em * (1 + 2.5*vf) ) 
        else:
            print('Error in calculation of E using Einstein equation.')
            print('Adhesion parameter can be only 0 or 1.')
            sys.exit()


class Guth:
    '''
    Guth equation for polymer composites = matrix with filler.
    
    * Guth equation is an extension of Einstein equation
      for composites with good interfacial adhesion.
    * Guth equation adds an correction for interparticle interactions
      in the form of additional term, which is a function of vf**2.
    * Assumption (like in Einstein eq,): Ef >> Em,
      i.e. the filler is much stiffer than the matrix.
    '''
    
    def E(Em,vf):
        '''
        Predict elastic modulus of a composite by means of Guth equation.

        Parameters
        ----------
        Em : float
            Elastic modulus of the polymer matrix.
        vf : float
            Volume fraction of the filler.

        Returns
        -------
        E : float
            Elastic modulus of the composite.
            
        Test
        ----
        >>> print(f'{Guth.E(1, 0.05):.3f}')
        1.160
        '''
        return( Em * (1 + 2.5*vf + 14.1*vf**2) )


class Kerner:
    '''
    Kerner equation for polymer composites = matrix with filler.
    
    * Kerner equation is an extension of Einstein equation
      for composites with good interfacial adhesion.
    * Kerner equation considers: (i) non-linear increase in modulus
      for higher filler concentrations and (ii) Poissons ratio of the matrix.
    * Assumption (like in Einstein eq,): Ef >> Em,
      i.e. the filler is much stiffer than the matrix.    
    '''

    def E(Em,vf, poisson = 0.4):
        '''
        Calculate modulus E of a polymer composite using Kerner equation.
        
        Parameters
        ----------
        Em : float
            Elastic modulus of the polymer matrix.
        vf : float
            Volume fraction of the filler.
        poisson : foat, optional, default is 0.4
            Poisson ratio ot the polymer matrix.

        Returns
        -------
        E : float
            Elastic modulus of the composite.
            
        
        Test
        ----
        >>> print(f'{Kerner.E(1, 0.05, poisson=0.4):.3f}')
        1.118
        '''
        E = Em * ( 1 + (15*(1-poisson))/(8-10*poisson) * vf/(1-vf) )
        return(E)



'''
Module: pmodels.ebm
-------------------

* Model: EBM -> system property = non-linear combination of component props.
* System: two-component polymer blend (isometric/non-oriented components).
* Predicted properties: elastic modulus (E) and yield strength (Y).

Simple usage

>>> # Standard import of pmodels package
>>> import pmodels as pm
>>>
>>> # (1) Estimate elastic modulus (E) of an isotropic binary polymer blend,
>>> # where moduli of the components are E1 = 1 GPa and E2 = 3 GPa,
>>> # and volume fraction of the second component is 0.5.
>>> E_ebm = pm.ebm.E(E1=1, E2=3, v2=0.5)
>>>
>>> # (2) Print the result
>>> print(f'Modulus of the blend: {E_ebm:.2f}')
Modulus of the blend: 1.70
'''

import numpy as np

# This module is used to suppress some specific numpy warnings
# which uselessly warn against situation that are excluded due to np.where
import warnings

#------------------------------------------------------------------------------

def vfractions(v2, v1cr = 0.156, v2cr = 0.156, q1 = 1.8, q2 = 1.8):
    '''
    Calculate volume fractions v1p,v2p,v1s,v2s,vp,vs by means of EBM model.
    
    Parameters
    ----------
    v2 : float
        Volume fraction of the 2nd component.
        The volume fraction of the 1st component
        is not defined as it can be calculated as v1 = (1 - v2).
        The 1st component is by convention (but not necessarily) the matrix.
    v1cr : float, optional, default is 0.156
        Critical volume fraction of the 1st component
        = when the 1st component starts to be continunous.
    v2cr : float, optional, default is 0.156
        Critical volume fraction of the 2nd component
        = when the 2nd component starts to be continunous.
    q1 : float, optional, default is 1.8
        Critical exponent of the 1st component,
        employed in the estimation of continuous volume fraction v1p.
    q2 : float, optional, default is 1.8
        Critical exponent of the 2nd component, 
        employed in the estimation of continuous volume fraction v2p.
        
    Returns
    -------
    vp, v1p, v2p, vs, v1s, v2s : floats
        v1p and vp2 = volume fraction of the 1st and 2nd phase, respectively,
        which is continuous.
        v1s and v2s = volume fraction of the 1st and 2nd phase, respectively,
        which is particulate.
        vp = v1p + v2p = sum of vol.fractions of continuous phases,
        representing paralel branch of the EBM model.
        vs = v1s + v2s = sum of vol.fractions of particulate phases,
        representing serial branch of the EBM model.
    
    Technical note
    --------------
    * This function (vfractions) is usually not called directly.
    * In most cases, it is just called by other functions of this module.

    Note to optional arguments
    --------------------------
    * The optional arguments are 
      the critical volume fractions (v1cr,v2cr) and critical exponents (q1,q2).
    * In the first approximation, they can be left at their default values
      (v1cr = v2cr = 0.156) and (q1 = q2 = 1.8).
    * The default values of are estimated from the percolation theory
      as described in https://doi.org/10.1002/pen.24805 and references therein.
    * Better estimates of v1cr and v2cr (i.e. the volume fractions when the
      first and second phase start to be continuous) can be estimated from the
      analysis of electron micrographs or from some other independt method.
    * The final values of v1cr, v2cr, q1 and q2 for given system
      can be obtained by fitting the EBM model to experimental data.

    Test
    ----
    >>> [round(e,3) for e in vfractions(0.5)]
    [0.398, 0.199, 0.199, 0.602, 0.301, 0.301]
    '''
    
    # (1) If parameter v2 is a numpy-array we need a special treatment!
    if type(v2) == np.ndarray:
        v1 = 1-v2
        # suppress warning about invalid/negative value in power
        # (zero value is never used due to np.where, but numpy warns anyway
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            v1p = np.where(v1 >= v1cr, ((v1-v1cr) / (1-v1cr))**q1, 0)
            v2p = np.where(v2 >= v2cr, ((v2-v2cr) / (1-v2cr))**q2, 0)
            
    # (2) If parameter v2 is a number, we use standard calculations...
    else:
        # Convert v2 to float in order to get 100% compatibility with numpy
        v2 = float(v2) 
        v1  = 1-v2
        v1p = ((v1-v1cr) / (1-v1cr))**q1 if v1 >= v1cr else 0
        v2p = ((v2-v2cr) / (1-v2cr))**q2 if v2 >= v2cr else 0
   
    # (3) The rest is the same for both numbers and np.arrays
    v1s = v1 - v1p
    v2s = v2 - v2p
    vp  = v1p + v2p
    vs  = v1s + v2s
    
    # (4) Return final values (this is also the same for numbers and np.arrays)
    return(vp,v1p,v2p,vs,v1s,v2s)
    

def E(E1,E2,v2, v1cr = 0.156, v2cr = 0.156, q1 = 1.8, q2 = 1.8):
    '''\
    Calculate elastic modulus, E, by means of EBM model.
    
    Parameters
    ----------
    E1 : float
        Elastic modulus of the 1st component.
    E2 : float
        Elastic modulus of the 2nd component.
    v2 : float
        Volume fraction of the 2nd component.
        The volume fraction of the 1st component
        is not defined as it can be calculated as v1 = (1 - v2).
        The 1st component is by convention (but not necessarily) the matrix.
    v1cr : float, optional, default is 0.156
        Critical volume fraction of the 1st component
        = when the 1st component starts to be continunous.
    v2cr : float, optional, default is 0.156
        Critical volume fraction of the 2nd component
        = when the 2nd component starts to be continunous.
    q1 : float, optional, default is 1.8
        Critical exponent of the 1st component,
        employed in the estimation of continuous volume fraction v1p.
    q2 : float, optional, default is 1.8
        Critical exponent of the 2nd component, 
        employed in the estimation of continuous volume fraction v2p.
     
    Returns
    -------
    E : float
        The modulus of the blend according to EBM model.
        
    Note to optional arguments
    --------------------------
    * The optional arguments are 
      the critical volume fractions (v1cr,v2cr) and critical exponents (q1,q2).
    * In the first approximation, they can be left at their default values
      (v1cr = v2cr = 0.156) and (q1 = q2 = 1.8).
    * The default values of are estimated from the percolation theory
      as described in https://doi.org/10.1002/pen.24805 and references therein.
    * Better estimates of v1cr and v2cr (i.e. the volume fractions when the
      first and second phase start to be continuous) can be estimated from the
      analysis of electron micrographs or from some other independt method.
    * The final values of v1cr, v2cr, q1 and q2 for given system
      can be obtained by fitting the EBM model to experimental data.
        
    Test
    ----
    >>> round(E(1,3,0.5), 3)
    1.699
    '''
    
    # (1) Calculate volume fractions vij
    v1 = 1 - v2
    vp,v1p,v2p,vs,v1s,v2s = vfractions(v2,v1cr,v2cr,q1,q2)
    
    # (2) Calculate elastic modulus, E, using vij
    if type(v2) in (int,float):
        # (2a) Argument v2 is a number:
        # Step 1 = standard calculation of E (Ok except for v1=0 and v2=0)
        # (if v1=0 then v2=v2p=1 => v1s=v2s=0 => division by zero at the end
        # (if v2=0 then v1=v1p=1 => v1s=v2s=0 => division by zero at the end
        try:
            E = E1*v1p + E2*v2p + vs**2 / ( v1s/E1 + v2s/E2)
        # Step 2 = if the calculation failed, treat special cases (v1=0, v2=0)
        except ZeroDivisionError:
            if v1 == 0: return(E2)
            if v2 == 0: return(E1)
    else:
        # (2b) Argument v2 is an np.array ...
        # Step 1 = standard calculation of E (Ok except for v1=0 and v2=0)
        # (Suppress warning about zero division + perform standard calculation
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            E = E1*v1p + E2*v2p + vs**2 / (v1s/E1 + v2s/E2) 
        # Step 2 = treat special cases (v1=0, v2=0)
        # (Numpy/vector calculations yields nan for v1=0 and v2=0
        E = np.where(v1 == 0, E2, E)  # correct E values for v1=0
        E = np.where(v2 == 0, E1, E)  # correct E values for v2=0
    
    # (3) Return final value (this is the same for np.arrays and numbers)
    return(E)


def Y(Y1,Y2,v2, v1cr = 0.156, v2cr = 0.156, q1 = 1.8, q2 = 1.8, A = 1):
    '''
    Calculate yield stress, Y, by means of EBM model.
    
    Parameters
    ----------
    E1 : float
        Elastic modulus of the 1st component.
    E2 : float
        Elastic modulus of the 2nd component.
    v2 : float
        Volume fraction of the 2nd component.
        The volume fraction of the 1st component
        is not defined as it can be calculated as v1 = (1 - v2).
        The 1st component is by convention (but not necessarily) the matrix.
    v1cr : float, optional, default is 0.156
        Critical volume fraction of the 1st component
        = when the 1st component starts to be continunous.
    v2cr : float, optional, default is 0.156
        Critical volume fraction of the 2nd component
        = when the 2nd component starts to be continunous.
    q1 : float, optional, default is 1.8
        Critical exponent of the 1st component,
        employed in the estimation of continuous volume fraction v1p.
    q2 : float, optional, default is 1.8
        Critical exponent of the 2nd component, 
        employed in the estimation of continuous volume fraction v2p.
    A : float, optional, default is 1
        Adhesion parameter, which ranges
        from A = 0 (zero interfacial adhesion)
        to A = 1 (perfect interfacial adhesion, no debonding before yield).

    Returns
    -------
    Y : float
        The yield stress of the blend according to EBM model.
        
    Note to optional arguments
    --------------------------
    * The optional arguments are 
      the critical volume fractions (v1cr,v2cr) and critical exponents (q1,q2).
    * In the first approximation, they can be left at their default values
      (v1cr = v2cr = 0.156) and (q1 = q2 = 1.8).
    * The default values of are estimated from the percolation theory
      as described in https://doi.org/10.1002/pen.24805 and references therein.
    * Better estimates of v1cr and v2cr (i.e. the volume fractions when the
      first and second phase start to be continuous) can be estimated from the
      analysis of electron micrographs or from some other independt method.
    * The final values of v1cr, v2cr, q1 and q2 for given system
      can be obtained by fitting the EBM model to experimental data.
        
    Test
    ----
    >>> round(Y(100,200,0.5), 3)
    119.879
    '''        
    
    # (1) Calculate volume fractions, vij
    vp,v1p,v2p,vs,v1s,v2s = vfractions(v2,v1cr,v2cr,q1,q2)
    
    # (2) Calculate yield stress, Y, using the volume fractions vij
    Y = Y1*v1p + Y2*v2p + A*min(Y1,Y2)*vs
    
    # (3) Return the calculated value
    return(Y)



'''
Package: PModels
----------------
* Collection of models to predict properties of polymer systems.
* The polymer systems = polymer blends and polymer composites.
* The final system properties
  are estimated from the properties of individual components. 

Ultra-brief example

>>> # Standard import of PModels package
>>> import pmodels as pm
>>> # Predict elastic modulus (E)
>>> # of a binary isotropic blend with composition 50/50
>>> # and moduli of the individual components E1 = 1 GPa and E2 = 3 GPa
>>> E_lin = pm.lin.P(1, 3, 0.5)  # linear model prediction   => E_lin = 2.00
>>> E_ebm = pm.ebm.E(1, 3, 0.5)  # EBM model, default params => E_ebm = 1.70

List of modules = predictive models

* pmodels.lin = LINear model
    - arbitrary systems, arbitrary property (P), linear prediction
* pmodels.ebm = Equivalent Box Model
    - isotropic polymer blends, modulus (E) or yield stress (Y), non-linear
* pmodels.ht = Halpin-Tsai equations
    - polymer composites with 0D/1D/2D fillers, modulus (E) 
* pmodels.comp = equations for polymer composites
    - polymer composites with fillers with/without interfacial adhesion
'''

__version__ = '0.4'

# The following command enables to use all submodules as follows:
# >>> from pmodels import *
# >>> E_lin = lin.P(1,3,0.5)
# >>> E_ebm = ebm.E(1,3,0.5)
__all__ = ['lin', 'ht', 'ebm', 'comp']

# The following block of commands enables to use all submodules as follows:
# >>> import pmodels as pm
# >>> E_lin = pm.lin.P(1,3,0.5)
# >>> E_ebm = pm.ebm.E(1,3,0.5)
import pmodels.lin
import pmodels.ht
import pmodels.ebm
import pmodels.comp



'''
Module: pmodels.ht
-------------------

* This module defines Halpin-Tsai equations.
* Halpin-Tsai equations predict elastic modulus of polymer composites.
* The polymer composite consists of a polymer matrix
  and stiff fibers or plates that are can be oriented or non-oriented.

Simple usage

>>> # Standard import of pmodels package
>>> import pmodels as pm
>>>
>>> # (1) Estimate elastic modulus (E) of a polymer composite,
>>> # with PA6 matrix (Em = 1 GPa) and 10% of oriented fibers (Ef = 10 GPa)
>>> # with three different aspect ratios (AR = 100, 10, and 1).
>>>
>>> Ec_001 = pm.ht.E(Em=1, Ef=10, vf=0.1, ftype=1, orient=1, AR=1)
>>> Ec_010 = pm.ht.E(Em=1, Ef=10, vf=0.1, ftype=1, orient=1, AR=10)
>>> Ec_100 = pm.ht.E(Em=1, Ef=10, vf=0.1, ftype=1, orient=1, AR=100)
>>>
>>> # (2) Print the result
>>> print(f'Elastic moduli: {Ec_001:.2f} {Ec_010:.2f} {Ec_100:.2f}')
Elastic moduli: 1.24 1.65 1.87
'''

#------------------------------------------------------------------------------

# Custom exceptions
class _UnknownFillerTypeException(Exception): pass
class _UnknownOrientationException(Exception): pass

# Main function: general calculation of elastic modulus according to HT theory
def E(Em,Ef,vf, ftype = 1, orient = 1, AR = 1):
    '''
    Predict elastic modulus, E, using Halpin-Tsai equations.
    
    Parameters
    ----------
    Em : float
        Elastic modulus of the matrix.
    Ef : float
        Elastic modulus of the filler.
    vf : float
        Volume fraction of the filler.
    ftype  : 1 or 2, optional, default is 1
        If ftype = 1 => the filler = fibers.
        If ftype = 2 => the filler = plates.
    orient : 1 or 2 or 3, optional, default is 1
        If orient = 1 => fibers/plates are parallel to deformation.
        If orient = 2 => fibers/plates are perpendicular to deformation.
        If orient = 3 => fibers/plates are oriented randomly.
    AR : float, optional, default is 1
        Aspect ratio = l/d for fibers, d/t for platelets
        (l = length, d = diameter, t = thickness).        
    
    Returns
    -------
    E : float
        The final modulus of the system according to HT model.
        
    Test
    ----
    >>> round(E(100,200,0.5,AR=1),3)
    142.857
    >>> round(E(100,200,0.5,AR=100),3)
    149.876
    '''
    
    # Local/internal functions
    def E_general_formula(ksi):
        eta = (Ef/Em - 1) / (Ef/Em + ksi)
        E = (1 + ksi*eta*vf) / (1 - eta*vf)
        return(E)
    def E_unoriented_fibers(): return(0.184*E_parallel + 0.816*E_perpendicular)
    def E_unoriented_plates(): return(0.490*E_parallel + 0.510*E_perpendicular)
    
    # Main function definition
    try:
        if ftype == 1:
            E_parallel      = Em * E_general_formula(ksi=2*AR)
            E_perpendicular = Em * E_general_formula(ksi=2)
        elif ftype == 2:
            E_parallel      = Em * E_general_formula(ksi=2*AR)
            E_perpendicular = Em * E_general_formula(ksi=2)
        else:
            raise _UnknownFillerTypeException()
        if   orient == 1: return(E_parallel)
        elif orient == 2: return(E_perpendicular)
        elif orient == 3:
            if   ftype == 1: return(E_unoriented_fibers())
            elif ftype == 2: return(E_unoriented_plates())
        else:
            raise _UnknownOrientationException()
    except _UnknownFillerTypeException:
        print('Unknown filler type!')
        print('Allowed: 1 = fibers, 2 = plates')
    except _UnknownOrientationException:
        print('Unknown orientation.')
        print('Allowed: 1 = parallel, 2 = perpendicular, 3 = random')
    
    # End of main function, return final value
    return(E)
